@@ -105,9 +105,9 @@ namespace DisplayPort
         NvU8     cachedSourceChipRevision;
         bool     bOuiCached;
 
-        unsigned ouiId;                                             // Sink ouiId
-        char modelName[NV_DPCD_SOURCE_DEV_ID_STRING__SIZE + 1];     // Device Model-name
-        bool    bIgnoreSrcOuiHandshake;                             // Skip writing source OUI
+        unsigned ouiId;                                                   // Sink ouiId
+        unsigned char modelName[NV_DPCD_SOURCE_DEV_ID_STRING__SIZE + 1];  // Device Model-name
+        bool    bIgnoreSrcOuiHandshake;                                   // Skip writing source OUI
 
         LinkPolicy    linkPolicy;
 
@@ -197,6 +197,8 @@ namespace DisplayPort
         unsigned compoundQueryLocalLinkPBN;
         NvU64 compoundQueryUsedTunnelingBw;
         bool compoundQueryForceEnableFEC;
+        // WAR
+        bool bDP2XPreferNonDSCForLowPClk;
 
         unsigned freeSlots;
         unsigned maximumSlots;
@@ -214,6 +216,7 @@ namespace DisplayPort
         // this is the link config requested by a client.
         // can be set and reset by the client for a given operation.
         LinkConfiguration preferredLinkConfig;
+        bool forcePreferredLinkConfig;
 
         //
         // Desired link configuration of single head multiple sst secondary connector.
@@ -261,7 +264,7 @@ namespace DisplayPort
         bool        bSkipLt;
 
         // Flag to make sure that zombie gets triggred when a powerChange event happens
-        bool bMitigateZombie;
+        bool        bMitigateZombie;
 
         //
         // HP Valor QHD+ N15P-Q3 EDP needs 50ms delay after D3
@@ -323,6 +326,9 @@ namespace DisplayPort
         //
         NvU32       LT2FecLatencyMs;
 
+        // On eDP, do not cache the source OUI if it reads 0. See bug 4793112
+        bool        bSkipZeroOuiCache;
+
         //
         // Dual SST Partner connector object pointer
         ConnectorImpl *pCoupledConnector;
@@ -360,13 +366,9 @@ namespace DisplayPort
         // the stale messages from previous discovery.
         //
         bool        bForceClearPendingMsg;
-        bool        bSkipFakeDeviceDpcdAccess;
-
         NvU64       allocatedDpTunnelBw;
         NvU64       allocatedDpTunnelBwShadow;
         bool        bForceDisableTunnelBwAllocation;
-        bool        bClientRequestedDpTunnelBwAllocation;
-        bool        bIsDpTunnelBwAllocationEnabled;
 
         Group *perHeadAttachedGroup[NV_MAX_HEADS];
         NvU32 inTransitionHeadMask;
@@ -378,17 +380,18 @@ namespace DisplayPort
         void setPolicyAssessLinkSafely(bool enabled);
 
         void discoveryDetectComplete();
-        void discoveryNewDevice(const DiscoveryManager::Device & device);
-        void discoveryLostDevice(const Address & address);
-        void processNewDevice(const DiscoveryManager::Device & device,
-            const Edid & edid,
+        void discoveryNewDevice(const DiscoveryManager::Device &device);
+        void discoveryLostDevice(const Address &address);
+        void processNewDevice(const DiscoveryManager::Device &device,
+            const Edid &edid,
             bool isMultistream,
             DwnStreamPortType portType,
             DwnStreamPortAttribute portAttribute,
             bool isCompliance = false);
 
-        void applyEdidWARs(Edid & edid, DiscoveryManager::Device device);
-        void applyRegkeyOverrides(const DP_REGKEY_DATABASE& dpRegkeyDatabase);
+        void applyEdidWARs(Edid &edid, DiscoveryManager::Device &device);
+        virtual void handleEdidWARs(Edid &edid, DiscoveryManager::Device &device){};
+        void applyRegkeyOverrides(const DP_REGKEY_DATABASE &dpRegkeyDatabase);
 
         ResStatusNotifyMessage ResStatus;
 
@@ -403,17 +406,20 @@ namespace DisplayPort
         virtual LinkConfiguration getMaxLinkConfig();
         virtual LinkConfiguration getActiveLinkConfig();
         virtual void powerdownLink(bool bPowerdownPanel = false);
+        LinkConfiguration initMaxLinkConfig();
 
         GroupImpl * getActiveGroupForSST();
         bool detectSinkCountChange();
-        bool handlePhyPatternRequest();
-        void applyOuiWARs();
+
+        virtual bool handlePhyPatternRequest();
+        virtual bool handleTestLinkTrainRequest();
+        virtual void applyOuiWARs();
         bool linkUseMultistream()
         {
             return (linkState == DP_TRANSPORT_MODE_MULTI_STREAM);
         }
 
-        void populateAllDpConfigs();
+        virtual void populateAllDpConfigs();
         virtual LinkRates* importDpLinkRates();
 
         //
@@ -565,24 +571,21 @@ namespace DisplayPort
         virtual void notifyAttachEnd(bool modesetCancelled);
         virtual void notifyDetachBegin(Group * target);
         virtual void notifyDetachEnd(bool bKeepOdAlive = false);
+        virtual bool willLinkSupportModeSST(const LinkConfiguration &linkConfig,
+                                            const ModesetInfo &modesetInfo,
+                                            const DscParams *pDscParams = NULL);
 
         bool performIeeeOuiHandshake();
         void setIgnoreSourceOuiHandshake(bool bIgnore);
         bool getIgnoreSourceOuiHandshake();
-        bool willLinkSupportModeSST(const LinkConfiguration & linkConfig, const ModesetInfo & modesetInfo);
         void forceLinkTraining();
 
         bool     updateDpTunnelBwAllocation();
-        void     configureDpTunnelBwAllocation();
         TriState requestDpTunnelBw(NvU8 requestedBw);
         bool     allocateDpTunnelBw(NvU64 bandwidth);
         bool     allocateMaxDpTunnelBw();
         NvU64    getMaxTunnelBw();
-
-        void enableDpTunnelingBwAllocationSupport()
-        {
-            bClientRequestedDpTunnelBwAllocation = true;
-        }
+        void     enableDpTunnelingBwAllocationSupport();
 
         void assessLink(LinkTrainingType trainType = NORMAL_LINK_TRAINING);
 
@@ -605,7 +608,8 @@ namespace DisplayPort
         }
         bool trainLinkOptimized(LinkConfiguration lConfig);
         bool trainLinkOptimizedSingleHeadMultipleSST(GroupImpl * group);
-        bool getValidLowestLinkConfig(LinkConfiguration & lConfig, LinkConfiguration & lowestSelected, ModesetInfo queryModesetInfo);
+        bool getValidLowestLinkConfig(LinkConfiguration &lConfig, LinkConfiguration &lowestSelected,
+                                      ModesetInfo queryModesetInfo, const DscParams *pDscParams = NULL);
         bool postLTAdjustment(const LinkConfiguration &, bool force);
         void populateUpdatedLaneSettings(NvU8* voltageSwingLane, NvU8* preemphasisLane, NvU32 *data);
         void populateDscCaps(DSC_INFO* dscInfo, DeviceImpl * dev, DSC_INFO::FORCED_DSC_PARAMS* forcedParams);
@@ -615,8 +619,8 @@ namespace DisplayPort
         void populateDscBranchCaps(DSC_INFO* dscInfo, DeviceImpl * dev);
         void populateDscModesetInfo(MODESET_INFO * pModesetInfo, const DpModesetParams * pModesetParams);
 
-        virtual bool train(const LinkConfiguration & lConfig, bool force, LinkTrainingType trainType = NORMAL_LINK_TRAINING);
-        virtual bool validateLinkConfiguration(const LinkConfiguration & lConfig);
+        virtual bool train(const LinkConfiguration &lConfig, bool force, LinkTrainingType trainType = NORMAL_LINK_TRAINING);
+        virtual bool validateLinkConfiguration(const LinkConfiguration &lConfig);
 
         virtual bool assessPCONLinkCapability(PCONLinkControl *params);
         bool trainPCONFrlLink(PCONLinkControl *pConControl);
@@ -625,7 +629,7 @@ namespace DisplayPort
         bool setDeviceDscState(Device * dev, bool bEnableDsc);
 
         // the lowest level function(nearest to the hal) for the connector.
-        bool rawTrain(const LinkConfiguration & lConfig, bool force, LinkTrainingType linkTrainingType);
+        bool rawTrain(const LinkConfiguration &lConfig, bool force, LinkTrainingType linkTrainingType);
 
         virtual bool enableFlush();
         virtual bool beforeAddStream(GroupImpl * group, bool force=false, bool forFlushMode = false);
@@ -640,11 +644,12 @@ namespace DisplayPort
 
         bool deleteAllVirtualChannels();
         void clearTimeslices();
+        virtual void applyTimeslotWAR(unsigned &slot_count){};
         virtual bool allocateTimeslice(GroupImpl * targetGroup);
         void freeTimeslice(GroupImpl * targetGroup);
         void flushTimeslotsToHardware();
         bool getHDCPAbortCodesDP12(NvU32 &hdcpAbortCodesDP12);
-        bool getOuiSink(unsigned &ouiId, char * modelName, size_t modelNameBufferSize, NvU8 & chipRevision);
+        bool getOuiSink(unsigned &ouiId, unsigned char * modelName, size_t modelNameBufferSize, NvU8 &chipRevision);
         bool hdcpValidateKsv(const NvU8 *ksv, NvU32 Size);
         void cancelHdcpCallbacks();
         bool handleCPIRQ();
@@ -653,10 +658,10 @@ namespace DisplayPort
         void handleDpTunnelingIrq();
         void handleHdmiLinkStatusChanged();
         void sortActiveGroups(bool ascending);
-        void configInit();
         void handlePanelReplayError();
 
-        virtual DeviceImpl* findDeviceInList(const Address & address);
+        virtual void configInit();
+        virtual DeviceImpl* findDeviceInList(const Address &address);
         virtual void disconnectDeviceList();
         void notifyLongPulseInternal(bool statusConnected);
         virtual void notifyLongPulse(bool status);
@@ -672,7 +677,10 @@ namespace DisplayPort
         virtual bool isFECSupported();
         virtual bool isFECCapable();
         virtual NvU32 maxLinkRateSupported();
-        virtual bool setPreferredLinkConfig(LinkConfiguration & lc, bool commit, bool force = false, LinkTrainingType trainType = NORMAL_LINK_TRAINING);
+        bool setPreferredLinkConfig(LinkConfiguration &lc, bool commit,
+                                    bool force = false,
+                                    LinkTrainingType trainType = NORMAL_LINK_TRAINING,
+                                    bool forcePreferredLinkConfig = false);
         virtual bool resetPreferredLinkConfig(bool force = false);
         virtual void setAllowMultiStreaming(bool bAllowMST);
         virtual bool getAllowMultiStreaming(void);
@@ -692,7 +700,29 @@ namespace DisplayPort
         bool setTestPattern(NV0073_CTRL_DP_TESTPATTERN testPattern, NvU8 laneMask, NV0073_CTRL_DP_CSTM cstm, NvBool bIsHBR2, NvBool bSkipLaneDataOverride = false);
         bool getLaneConfig(NvU32 *numLanes, NvU32 *data);    // "data" is an array of NV0073_CTRL_MAX_LANES unsigned ints
         bool setLaneConfig(NvU32 numLanes, NvU32 *data);    // "data" is an array of NV0073_CTRL_MAX_LANES unsigned ints
-        void getCurrentLinkConfig(unsigned & laneCount, NvU64 & linkRate);  // CurrentLink Configuration
+        // "data" is an array of NV0073_CTRL_MAX_LANES unsigned ints
+        virtual bool getDp2xLaneConfig(NvU32 *numLanes, NvU32 *data)
+        {
+            // The function is only for Connector class 2x (or later) that supports 128b/132b channel coding.
+            return false;
+        }
+        // "data" is an array of NV0073_CTRL_MAX_LANES unsigned ints
+        virtual bool setDp2xLaneConfig(NvU32 numLanes, NvU32 *data)
+        {
+            // The function is only for Connector class 2x (or later) that supports 128b/132b channel coding.
+            return false;
+        }
+
+        virtual void setDisableDownspread(bool _bDisableDownspread)
+        {
+            return;
+        }
+        virtual bool getDownspreadDisabled()
+        {
+            return false;
+        }
+
+        void getCurrentLinkConfig(unsigned &laneCount, NvU64 &linkRate);  // CurrentLink Configuration
         unsigned getPanelDataClockMultiplier();
         unsigned getGpuDataClockMultiplier();
         void configurePowerState(bool bPowerUp);
