@@ -86,8 +86,8 @@ _knvlinkFilterDiscoveredLinks
         pKernelNvlink->discoveredLinks &= ~(pKernelNvlink->vbiosDisabledLinkMask);
 
         NV_PRINTF(LEVEL_INFO,
-                  "Links discovered after VBIOS overrides = 0x%x\n",
-                  pKernelNvlink->discoveredLinks);
+                  "Links discovered after VBIOS overrides = 0x%llx\n",
+                  KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 64));
     }
 
     // Filter links that are disabled through regkey overrides
@@ -96,8 +96,8 @@ _knvlinkFilterDiscoveredLinks
         pKernelNvlink->discoveredLinks &= ~(pKernelNvlink->regkeyDisabledLinksMask);
 
         NV_PRINTF(LEVEL_INFO,
-                  "Links after applying disable links regkey = 0x%x\n",
-                  pKernelNvlink->discoveredLinks);
+                  "Links after applying disable links regkey = 0x%llx\n",
+                  KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 64));
     }
 
     return NV_OK;
@@ -135,7 +135,7 @@ _knvlinkFilterIoctrls
         NvU32 localDiscoveredLinks = kioctrlGetLocalDiscoveredLinks(pGpu, pKernelIoctrl);
 
         localDiscoveredLinks &=
-            kioctrlGetGlobalToLocalMask(pGpu, pKernelIoctrl, pKernelNvlink->discoveredLinks);
+            kioctrlGetGlobalToLocalMask(pGpu, pKernelIoctrl, KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 32));
 
         // No need to handle the IOCTRL if no links are being enabled
         if (localDiscoveredLinks == 0x0)
@@ -237,6 +237,7 @@ knvlinkConstructEngine_IMPL
         // Mark recovery info as valid for use by error recovery workqueues
         pInfo->bValid = NV_TRUE;
         pInfo->DomainBusDevice = gpuGetDBDF(pGpu);
+        pInfo->active = 0;
     }
 
     return NV_OK;
@@ -355,7 +356,7 @@ knvlinkStateLoad_IMPL
     KernelMIGManager *pKernelMIGManager = GPU_GET_KERNEL_MIG_MANAGER(pGpu);
     NvBool            bMIGNvLinkP2PDisabled = ((pKernelMIGManager != NULL) &&
                                                !kmigmgrIsMIGNvlinkP2PSupported(pGpu, pKernelMIGManager));
-    NvU32             preInitializedLinks;
+    NvU64             preInitializedLinks;
     NvU32             i;
     OBJTMR            *pTmr = GPU_GET_TIMER(pGpu);
 
@@ -440,10 +441,10 @@ knvlinkStateLoad_IMPL
     }
 
     // Remove the init disabled links from the discovered links mask
-    pKernelNvlink->discoveredLinks &= ~(pKernelNvlink->initDisabledLinksMask);
+    pKernelNvlink->discoveredLinks &= ~((NvU64)pKernelNvlink->initDisabledLinksMask);
 
     // Track un-connected links, we assume all discovered links are connected.
-    pKernelNvlink->connectedLinksMask = pKernelNvlink->discoveredLinks;
+    pKernelNvlink->connectedLinksMask = KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 32);
 
     // Initialize initializedLinks to 0 (assume no links initialized)
     pKernelNvlink->initializedLinks = 0;
@@ -497,8 +498,8 @@ knvlinkStateLoad_IMPL
     //
     if (pKernelNvlink->bRegistryLinkOverride)
     {
-        pKernelNvlink->enabledLinks = pKernelNvlink->discoveredLinks &
-                                      pKernelNvlink->registryLinkMask;
+        pKernelNvlink->enabledLinks = KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 64) &
+                                      KNVLINK_GET_MASK(pKernelNvlink, registryLinkMask, 64);
     }
     else if (bMIGNvLinkP2PDisabled)
     {
@@ -507,21 +508,21 @@ knvlinkStateLoad_IMPL
     }
     else
     {
-        pKernelNvlink->enabledLinks = pKernelNvlink->discoveredLinks;
+        pKernelNvlink->enabledLinks = KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 64);
     }
 
     // Sense NVLink bridge presence and remove links on missing bridges.
     knvlinkFilterBridgeLinks_HAL(pGpu, pKernelNvlink);
 
     // Disconnected links mask tracks links whose remote ends are not discovered
-    pKernelNvlink->disconnectedLinkMask = pKernelNvlink->enabledLinks;
+    pKernelNvlink->disconnectedLinkMask = KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32);
 
     if (!IS_RTLSIM(pGpu) || pKernelNvlink->bForceEnableCoreLibRtlsims)
     {
         if (!knvlinkPoweredUpForD3_HAL(pGpu, pKernelNvlink))
         {
             // Register links in the nvlink core library
-            FOR_EACH_INDEX_IN_MASK(32, i, pKernelNvlink->enabledLinks)
+            FOR_EACH_INDEX_IN_MASK(64, i, KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 64))
             {
                 status = knvlinkCoreAddLink(pGpu, pKernelNvlink, i);
                 if (status != NV_OK)
@@ -635,7 +636,7 @@ knvlinkStateLoad_IMPL
                                      PDB_PROP_KNVLINK_MINION_FORCE_ALI_TRAINING)))
     {
         status = knvlinkPreTrainLinksToActiveAli(pGpu, pKernelNvlink,
-                                                 pKernelNvlink->enabledLinks, NV_TRUE);
+                                                 KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32), NV_TRUE);
         if (status != NV_OK)
         {
             goto knvlinkStateLoad_end;
@@ -647,7 +648,7 @@ knvlinkStateLoad_IMPL
         // will be queries via DLSTAT to know their status and training
         // progression.
         //
-        FOR_EACH_INDEX_IN_MASK(32, i, pKernelNvlink->enabledLinks)
+        FOR_EACH_INDEX_IN_MASK(32, i, KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32))
         {
             status = knvlinkTrainLinksToActiveAli(pGpu, pKernelNvlink, NVBIT(i), NV_FALSE);
             if (status != NV_OK)
@@ -671,7 +672,7 @@ knvlinkStateLoad_IMPL
         FOR_EACH_INDEX_IN_MASK_END;
     }
 
-    FOR_EACH_INDEX_IN_MASK(32, i, pKernelNvlink->enabledLinks)
+    FOR_EACH_INDEX_IN_MASK(32, i, KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32))
     {
         status = tmrEventCreate(pTmr, &pKernelNvlink->nvlinkLinks[i].pTmrEvent,
                             ioctrlFaultUpTmrHandler, NULL,
@@ -921,14 +922,14 @@ knvlinkStatePostUnload_IMPL
                         !pSys->getProperty(pSys, PDB_PROP_SYS_FABRIC_IS_EXTERNALLY_MANAGED) &&
                         !API_GPU_IN_RESET_SANITY_CHECK(pGpu))
     {
-        NV2080_CTRL_NVLINK_DISABLE_DL_INTERRUPTS_PARAMS params;
+        NV2080_CTRL_INTERNAL_NVLINK_DISABLE_DL_INTERRUPTS_PARAMS params;
 
         portMemSet(&params, 0, sizeof(params));
-        params.linkMask = pKernelNvlink->enabledLinks;
+        params.linkMask = KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32);
 
         // Disable all the DL interrupts
         status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
-                                     NV2080_CTRL_CMD_NVLINK_DISABLE_DL_INTERRUPTS,
+                                     NV2080_CTRL_CMD_INTERNAL_NVLINK_DISABLE_DL_INTERRUPTS,
                                      (void *)&params, sizeof(params));
         if (status != NV_OK)
         {
@@ -948,7 +949,7 @@ knvlinkStatePostUnload_IMPL
     }
 
 #if defined(INCLUDE_NVLINK_LIB)
-    FOR_EACH_INDEX_IN_MASK(32, linkId, pKernelNvlink->enabledLinks)
+    FOR_EACH_INDEX_IN_MASK(32, linkId, KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32))
     {
         // Update remote GPU disconnectedLinkMasks
         OBJGPU *pRemoteGpu = gpumgrGetGpuFromBusInfo(pKernelNvlink->nvlinkLinks[linkId].remoteEndInfo.domain,
@@ -998,7 +999,7 @@ _knvlinkPurgeState
     NvBool bMIGNvLinkP2PDisabled = ((pKernelMIGManager != NULL) &&
                           !kmigmgrIsMIGNvlinkP2PSupported(pGpu, pKernelMIGManager));
 
-    FOR_EACH_INDEX_IN_MASK(32, linkId, pKernelNvlink->discoveredLinks)
+    FOR_EACH_INDEX_IN_MASK(32, linkId, KNVLINK_GET_MASK(pKernelNvlink, discoveredLinks, 32))
     {
         if ((pKernelNvlink->nvlinkLinks[linkId].pTmrEvent != NULL) && (pTmr != NULL))
         {
@@ -1032,7 +1033,7 @@ _knvlinkPurgeState
         if (pKernelNvlink->pNvlinkDev)
         {
             // Un-register the links from nvlink core library
-            FOR_EACH_INDEX_IN_MASK(32, linkId, pKernelNvlink->enabledLinks)
+            FOR_EACH_INDEX_IN_MASK(32, linkId, KNVLINK_GET_MASK(pKernelNvlink, enabledLinks, 32))
             {
                 knvlinkCoreRemoveLink(pGpu, pKernelNvlink, linkId);
             }
@@ -1101,14 +1102,41 @@ knvlinkShutdownLinks_WORKITEM
 {
     OBJGPU *pGpu                = gpumgrGetGpu(gpuInstance);
     KernelNvlink *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NV_STATUS status;
 
     // Sanity Checks
     NV_CHECK_OR_RETURN_VOID(LEVEL_INFO,
             (pGpu != NULL &&
              pKernelNvlink != NULL));
 
-    NV_ASSERT_OR_RETURN_VOID(
+    if (pKernelNvlink->ipVerNvlink >= NVLINK_VERSION_50)
+    {
+        // Unilaterally shut down all enabled links
+        NV2080_CTRL_NVLINK_PROCESS_INIT_DISABLED_LINKS_PARAMS disableLinksParams = { 0 };
+        disableLinksParams.initDisabledLinksMask = pKernelNvlink->enabledLinks;
+        status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
+                        NV2080_CTRL_CMD_NVLINK_PROCESS_INIT_DISABLED_LINKS,
+                        (void *)&disableLinksParams,
+                        sizeof(NV2080_CTRL_NVLINK_PROCESS_INIT_DISABLED_LINKS_PARAMS));
+
+        if (status != NV_OK)
+        {
+            NV_PRINTF(LEVEL_ERROR, "Failed to send disable link RPC on degraded GPU%d\n",
+                pGpu->gpuInstance);
+        }
+
+        // Suspend and invalidate probe on degraded GPU.
+        if (gpuFabricProbeIsSupported(pGpu))
+        {
+            gpuFabricProbeSuspend(pGpu->pGpuFabricProbeInfoKernel);
+            gpuFabricProbeInvalidate(pGpu->pGpuFabricProbeInfoKernel);
+        }
+    }
+    else
+    {
+        NV_ASSERT_OR_RETURN_VOID(
             knvlinkCoreShutdownDeviceLinks(pGpu, pKernelNvlink, NV_TRUE) == NV_OK);
+    }
 }
 
 /*!
@@ -1126,14 +1154,17 @@ knvlinkSetDegradedMode_IMPL
 (
     OBJGPU       *pGpu,
     KernelNvlink *pKernelNvlink,
-    NvU32         linkId
+    NvS32         linkId
 )
 {
-    NvU32         gpuInstance;
-    OBJGPU       *pRemoteGpu = NULL;
-    KernelNvlink *pRemoteKernelNvlink = NULL;
+    NvU32 gpuInstance;
+    NvBool bShutdownLinks = NV_FALSE;
+    NV_STATUS status = NV_OK;
 
-    if (!pKernelNvlink)
+    // Prevent invalid shift later
+    NV_ASSERT_OR_RETURN_VOID((linkId == -1) || ((linkId >= 0) && (linkId < 32)));
+
+    if (pKernelNvlink == NULL)
     {
         NV_PRINTF(LEVEL_ERROR,
                 "Failed to get Local Nvlink info for linkId %d to update Degraded GPU%d status\n",
@@ -1142,63 +1173,70 @@ knvlinkSetDegradedMode_IMPL
         return;
     }
 
-    if(pKernelNvlink->bIsGpuDegraded)
+    if (pKernelNvlink->bIsGpuDegraded)
     {
         return;
     }
 
-    //Find the remote GPU/NVLink attached to this link, if any
+    // Find the remote GPU/NVLink attached to this link, if any
     for (gpuInstance = 0; gpuInstance < NV_MAX_DEVICES; gpuInstance++)
     {
-        if (pKernelNvlink->peerLinkMasks[gpuInstance] & NVBIT(linkId))
+        OBJGPU *pRemoteGpu = gpumgrGetGpu(gpuInstance);
+        KernelNvlink *pRemoteKernelNvlink;
+        NvU32 linkMask = (linkId == -1) ? pKernelNvlink->enabledLinks : NVBIT32(linkId);
+
+        if ((pKernelNvlink->peerLinkMasks[gpuInstance] & linkMask) == 0x0)
+            continue;
+
+        // Sanity checks
+        if (pRemoteGpu == NULL)
         {
-            pRemoteGpu = gpumgrGetGpu(gpuInstance);
-            break;
+            NV_PRINTF(LEVEL_ERROR,
+                    "Failed to get Remote GPU info for linkId %d to update Degraded GPU%d status\n",
+                    linkId, pGpu->gpuInstance);
+
+            continue;
+        }
+
+        pRemoteKernelNvlink = GPU_GET_KERNEL_NVLINK(pRemoteGpu);
+        if (pRemoteKernelNvlink == NULL)
+        {
+            NV_PRINTF(LEVEL_ERROR,
+                    "Failed to get Remote Nvlink info for linkId %d to update Degraded GPU%d status\n",
+                    linkId, pGpu->gpuInstance);
+
+            continue;
+        }
+
+        //
+        // Only mark the current GPU if the peer GPU is not degraded.
+        // In the case where the peer GPU is already degraded then there is no additional
+        // loss of connectivity in the fabric due to the link error that caused this
+        // path to be triggered.
+        //
+        if (!pRemoteKernelNvlink->bIsGpuDegraded)
+        {
+            pKernelNvlink->bIsGpuDegraded = NV_TRUE;
+            bShutdownLinks = NV_TRUE;
+            NV_PRINTF(LEVEL_ERROR,
+                    "GPU%d marked Degraded. Error originated on linkId %d!\n",
+                    pGpu->gpuInstance, linkId);
+
+
         }
     }
 
-    // Sanity checks
-    if (pRemoteGpu == NULL)
+    if (bShutdownLinks)
     {
-        NV_PRINTF(LEVEL_ERROR,
-                "Failed to get Remote GPU info for linkId %d to update Degraded GPU%d status\n",
-                linkId, pGpu->gpuInstance);
-
-        return;
-    }
-
-    pRemoteKernelNvlink = GPU_GET_KERNEL_NVLINK(pRemoteGpu);
-    if (!pRemoteKernelNvlink)
-    {
-        NV_PRINTF(LEVEL_ERROR,
-                "Failed to get Remote Nvlink info for linkId %d to update Degraded GPU%d status\n",
-                linkId, pGpu->gpuInstance);
-
-        return;
-    }
-
-    //
-    // Only mark the current GPU if the peer GPU is not degraded.
-    // In the case where the peer GPU is already degraded then there is no additional
-    // loss of connectivity in the fabric due to the link error that caused this
-    // path to be triggered.
-    //
-    if (pRemoteKernelNvlink->bIsGpuDegraded == NV_FALSE)
-    {
-        pKernelNvlink->bIsGpuDegraded = NV_TRUE;
-        NV_PRINTF(LEVEL_ERROR,
-                "GPU%d marked Degraded. Error originated on linkId %d!\n",
-                pGpu->gpuInstance, linkId);
-
         // Queue a workitem to handle rest of the degraded mode handling
-        NV_CHECK_OR_RETURN_VOID(LEVEL_ERROR, NV_OK ==
-                osQueueWorkItemWithFlags(pGpu, knvlinkShutdownLinks_WORKITEM,
-                                            NULL,
-                                            (OS_QUEUE_WORKITEM_FLAGS_LOCK_API_RO |
-                                                OS_QUEUE_WORKITEM_FLAGS_LOCK_GPU_GROUP_SUBDEVICE)));
-    }
 
-    return;
+        NV_CHECK_OK_OR_ELSE(status, LEVEL_ERROR,
+            osQueueWorkItemWithFlags(pGpu, knvlinkShutdownLinks_WORKITEM,
+                                     NULL,
+                                     (OS_QUEUE_WORKITEM_FLAGS_LOCK_API_RO |
+                                      OS_QUEUE_WORKITEM_FLAGS_LOCK_GPU_GROUP_SUBDEVICE)),
+            return;);
+    }
 }
 
 /*!
@@ -1219,6 +1257,20 @@ knvlinkGetDegradedMode_IMPL
     return pKernelNvlink->bIsGpuDegraded;
 }
 
+/*!
+ * @brief Is NVLINK Uncontained error recovery active?
+ */
+NvBool
+knvlinkIsUncontainedErrorRecoveryActive_IMPL
+(
+    OBJGPU       *pGpu,
+    KernelNvlink *pKernelNvlink
+)
+{
+    NVLINK_UNCONTAINED_ERROR_RECOVERY_INFO *pInfo = gpumgrGetNvlinkRecoveryInfo(gpuGetDBDF(pGpu));
+    return ((pInfo != NULL) && pInfo->bValid && (portAtomicOrU32(&pInfo->active, 0) != 0));
+}
+
 void
 knvlinkDestruct_IMPL
 (
@@ -1328,15 +1380,15 @@ _knvlinkProcessSysmemLinks
     }
 
     // Set Buffer ready for the sysmem links
-    NV2080_CTRL_NVLINK_PROGRAM_BUFFERREADY_PARAMS programBufferRdyParams;
+    NV2080_CTRL_INTERNAL_NVLINK_PROGRAM_BUFFERREADY_PARAMS programBufferRdyParams;
 
     portMemSet(&programBufferRdyParams, 0, sizeof(programBufferRdyParams));
-    programBufferRdyParams.flags        = NV2080_CTRL_NVLINK_PROGRAM_BUFFERREADY_FLAGS_SET;
+    programBufferRdyParams.flags        = NV2080_CTRL_INTERNAL_NVLINK_PROGRAM_BUFFERREADY_FLAGS_SET;
     programBufferRdyParams.bSysmem      = NV_TRUE;
     programBufferRdyParams.peerLinkMask = 0;
 
     status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
-                                 NV2080_CTRL_CMD_NVLINK_PROGRAM_BUFFERREADY,
+                                 NV2080_CTRL_CMD_INTERNAL_NVLINK_PROGRAM_BUFFERREADY,
                                  (void *)&programBufferRdyParams,
                                  sizeof(programBufferRdyParams));
     if (status != NV_OK)
@@ -1361,11 +1413,11 @@ _knvlinkProcessSysmemLinks
         // PPC+GV100.
         //
 
-        NV2080_CTRL_NVLINK_ENABLE_SYSMEM_NVLINK_ATS_PARAMS sysmemNvlinkAtsParams;
+        NV2080_CTRL_INTERNAL_NVLINK_ENABLE_SYSMEM_NVLINK_ATS_PARAMS sysmemNvlinkAtsParams;
         portMemSet(&sysmemNvlinkAtsParams, 0, sizeof(sysmemNvlinkAtsParams));
 
         status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
-                                     NV2080_CTRL_CMD_NVLINK_ENABLE_SYSMEM_NVLINK_ATS,
+                                     NV2080_CTRL_CMD_INTERNAL_NVLINK_ENABLE_SYSMEM_NVLINK_ATS,
                                      (void *)&sysmemNvlinkAtsParams,
                                      sizeof(sysmemNvlinkAtsParams));
         if (status != NV_OK)
@@ -1376,8 +1428,6 @@ _knvlinkProcessSysmemLinks
 
         status = kmemsysSetupCoherentCpuLink(pGpu, GPU_GET_KERNEL_MEMORY_SYSTEM(pGpu), bFlush);
         NV_ASSERT_OR_RETURN(status == NV_OK, status);
-
-        osSetNVLinkSysmemLinkState(pGpu, NV_TRUE);
     }
 
     return status;
